# *DocGen* is an reverse parser, or unparser. A parser converts text into an 
# structured internal representation. *DocGen* does the opposite. It converts 
# an structured internal representation into text. In other words, *DocGen* 
# converts data from data structures into text.

# Copyright (c) 2011-2015 by Designer's Guide Consulting, Inc.
# This code and documentation is the property of Designer's Guide Consulting and 
# is a corporate secret.  If you gain access to this code or documentation and 
# are not an employee of Designer's Guide Consulting, please notify 
# consulting@designers-guide.com immediately.  Until you hear back from them 
# please keep the code secret and avoid examining it. If you cannot reach them, 
# please destroy the code and documentation in a secure manner.

from .production import (
    Production,
    newlineCode, tabCode, killCode, emptyCode,
    indentCode, leaderCode, dedentCode, render
)
from .predefined import (
    Alternates, Block, Cull, DiscardIfEmpty, Indent, Indented, Join, Leader, 
    Lines, Newline, Op, Optional, ParenList, Required, StrikeLine, Tab
)
from .templet import Templet, TempletError
from .template import Template
